# -*- coding: utf-8 -*-
# cython: profile=False

#  Copyright © 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=C0326,W0614
# although this file is compiled with Python3 syntax, Cython needs at least division from __future__
from __future__ import print_function, division

# This line makes cython happy
global __name__, __package__    # pylint: disable=W0604
#px/__compiled = True
__compiled = False

#pxd/from qt cimport *
from .debug_purepython import *

#pxh>#include <QtCore/QString>
#pxh>QString gettext_translate(const char *text, const char *disambiguation);
#pxh>#include "../../data/ui/qt/main.h"
#pxh>#include "../../data/ui/qt/preferences.h"
#pxh>#include "../../data/ui/qt/help.h"
#pxh>#include "../../data/ui/qt/about.h"
#pxh>#include <QtCore/QPropertyAnimation>
#pxh>#include <QtGui/QDesktopServices>
#pxh>#include <QtGui/QKeyEvent>
#pxh>#include <QtGui/QStandardItemModel>
#pxh>#include <QtGui/QIconEngine>
#pxh>#include <QtGui/QPainter>
#pxh>#include <QtWidgets/QStyledItemDelegate>
#pxh>#include <QtWidgets/QWidget>
#pxh>#include <QtWidgets/QOpenGLWidget>
#pxh>#include <QtWidgets/QLineEdit>
#pxh>#include <QtWidgets/QColorDialog>
#pxh>#include <QtWidgets/QScrollBar>
#pxh>#include <QtWidgets/QListWidget>
#pxh>#include <QtWidgets/QListWidgetItem>
#pxh>#include <QtWidgets/QMenu>
#pxh>
#pxc>#include "[[_qtui_VARIANT]]_moc.h"
#pxc>#include "qtui_p.h"
#pxc>

#px/from libc.stdio cimport printf
def printf(fmt, *args): print(fmt % args, end='')
#px/from libc.stdio cimport puts
puts = print

#px/cdef enum: #
if True:
    DEBUG_MSGEXT = 0x1
    DEBUG_DRAW = 0x2
    DEBUG_FPS = 0x4
#px+cdef long debug
debug = 0
    
def set_debug_flags(module):
    global debug
    if module.DEBUG_MSGEXT: debug |= DEBUG_MSGEXT
    if module.DEBUG_DRAW: debug |= DEBUG_DRAW
    if module.DEBUG_FPS: debug |= DEBUG_FPS
    
#pxm-FUNC PD
def init_module():
    if debug & DEBUG_MSGEXT:
        print('init module:', __name__)
        print('  from package:', __package__)
        print('  compiled:', __compiled)
        
#px+cdef extern from "[[_qtui_VARIANT]]_moc.h":#
#pxm>INDENT

#px+cdef QString gettext_translate(const char *text, const char *disambiguation) with gil


#pxm-QC_CPPCLASS class ShortcutEditor (QLabel):
class ShortcutEditor (QLabel):
    #pxh>QString *key;
    #pxh>
    #pxm-FUNC CH
    def __init__(parent:'QWidget *'):  _("Press a key …")
    #pxc>    : QLabel(gettext_translate("Press a key …", Q_NULLPTR), parent)
        #pxc>{
        #pxc>    setFocusPolicy(Qt::StrongFocus);
        #pxc>    setAutoFillBackground(true);
        #pxc>}
        #pxc>
    #pxh>Q_SIGNAL void key_pressed();
        #pxh>
    #pxm-FUNC CH
    def keyPressEvent(event:'QKeyEvent *'):  pass
        #pxc>{
        #pxc>    switch (event->key()) {
        #pxc>        case Qt::Key_Shift: case Qt::Key_Control: case Qt::Key_Meta:
        #pxc>        case Qt::Key_Alt: case Qt::Key_AltGr:
        #pxc>        case Qt::Key_CapsLock: case Qt::Key_NumLock: case Qt::Key_ScrollLock:
        #pxc>            QLabel::keyPressEvent(event);
        #pxc>            return;
        #pxc>    }
        #pxc>    if (event->count() != 1) {
        #pxc>        QLabel::keyPressEvent(event);
        #pxc>        return;
        #pxc>    }
        #pxc>    int mod = event->modifiers() & (Qt::ShiftModifier | Qt::ControlModifier);
        #pxc>    QStringList keylist = QKeySequence(event->key() | int(mod)).toString().split('+');
        #pxc>    if ((event->modifiers() & Qt::KeypadModifier) && !keylist.contains(QStringLiteral("Num")))
        #pxc>        keylist.insert(keylist.size()-1, "Num");
        #pxc>    key = new QString(keylist.join('+'));
        #pxc>    key_pressed();
        #pxc>}
#pxm>CPPCLASS_END

#pxm-QC_CPPCLASS class ShortcutDelegate (QStyledItemDelegate):
class ShortcutDelegate (QStyledItemDelegate):
    #pxh>ShortcutDelegate(QObject *parent) : QStyledItemDelegate(parent) {}
    #pxh>
    #pxm-FUNC CH const
    def createEditor(parent:'QWidget *', option:'const QStyleOptionViewItem &', index:'const QModelIndex &')->'QWidget *':  pass
        #pxc>{
        #pxc>    ShortcutEditor *editor = new ShortcutEditor(parent);
        #pxc>    connect(editor, SIGNAL(key_pressed()), this, SLOT(on_editor_key_pressed()));
        #pxc>    return editor;
        #pxc>}
        #pxc>
    #pxh>void setEditorData(QWidget *editor, const QModelIndex &index) const {}
        #pxc>
    #pxm-FUNC CH const
    def setModelData(editor:'QWidget *', model:'QAbstractItemModel *', index:'const QModelIndex &'):  pass
        #pxc>{
        #pxc>    if (((ShortcutEditor*)editor)->key != NULL) {
        #pxc>        QVariant value(*((ShortcutEditor*)editor)->key);
        #pxc>        model->setData(index, value, Qt::DisplayRole);
        #pxc>    }
        #pxc>}
        #pxc>
#pxh>
    #pxm-FUNC CH
    def slot_on_editor_key_pressed():  pass
        #pxc>{
        #pxc>    QWidget *editor = (QWidget*)this->sender();
        #pxc>    this->commitData(editor);
        #pxc>    this->closeEditor(editor, QAbstractItemDelegate::NoHint);
        #pxc>}
#pxm>CPPCLASS_END

#pxm-C_CPPCLASS class ColorIconEngine (QIconEngine):
class ColorIconEngine (QIconEngine):
    #pxh>QString color;
    #pxh>
    #pxh>ColorIconEngine() : QIconEngine(), color(QStringLiteral("black")) {}
    #pxh>
    #pxm-FUNC CH
    def paint(painter:'QPainter *', rect:'const QRect &', mode:'QIcon::Mode', state:'QIcon::State'):  pass
        #pxc>{ painter->fillRect(rect, QColor(this->color)); }
    #pxm-FUNC CH const
    def clone()->'ColorIconEngine *':  pass
        #pxc>{
        #pxc>    ColorIconEngine *cloned = new ColorIconEngine();
        #pxc>    cloned->color = color;
        #pxc>    return cloned;
        #pxc>}
#pxm>CPPCLASS_END

#pxm-CPPCLASS
class UIContainer:
    #pxh>Ui::MainWindow ui_main;
    #pxh>QLabel *label_selectmodel;
    #pxh>int active_plugin_group;
    #pxh>Ui::DialogPreferences ui_pref;
    #pxh>QStandardItemModel *liststore_movekeys;
    #pxh>QStandardItemModel *liststore_faces;
    #pxh>int liststore_faces_width;
    #pxh>ColorIconEngine *iconengine_color;
    #pxh>ColorIconEngine *iconengine_background_color;
    #pxh>Ui::DialogHelp ui_help;
    #pxh>Ui::AboutDialog ui_about;
    #pxh>int index_tab_about;
    #pxh>int index_tab_license;
    #pxh>QScrollBar *scrollbar;
    #pxh>QPropertyAnimation animation;
    #pxh>
    #pxm-FUNC CHP nogil
    def setupUi_main(window:'QMainWindow *', icon1:1, icon2:1, icon3:'const QIcon &', key1:1, key2:'const QKeySequence &'):  pass
        #pxc>{
            #pxc>// actions that belongs to no widget
            #pxc>QAction *action_jump_to_editbar = new QAction(window);
            #pxc>action_jump_to_editbar->setObjectName("action_jump_to_editbar");
            #pxc>action_jump_to_editbar->setShortcut(key1);
            #pxc>window->addAction(action_jump_to_editbar);
            #pxc>QAction *action_edit_cube = new QAction(window);
            #pxc>action_edit_cube->setObjectName("action_edit_cube");
            #pxc>action_edit_cube->setShortcut(key2);
            #pxc>window->addAction(action_edit_cube);
            
            #pxc>// Qt Designer code
            #pxc>ui_main.setupUi(window);
            
            #pxc>// ...
            #pxc>ui_main.splitter->setCollapsible(0, false);
            #pxc>ui_main.splitter->setStretchFactor(0, 1);
            #pxc>ui_main.splitter->setStretchFactor(1, 0);
            #pxc>ui_main.listwidget->setItemDelegate(new ModelSelectionItemDelegate());
            
            #pxc>// set icons
            #pxc>ui_main.button_edit_exec->setIcon(QIcon::fromTheme("system-run"));
            #pxc>ui_main.button_edit_clear->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_main.action_challenge->setIcon(icon1);
            #pxc>ui_main.action_new_solved->setIcon(icon2);
            #pxc>ui_main.action_quit->setIcon(QIcon::fromTheme("application-exit"));
            #pxc>ui_main.action_rewind->setIcon(QIcon::fromTheme("media-seek-backward"));
            #pxc>ui_main.action_previous->setIcon(QIcon::fromTheme("media-skip-backward"));
            #pxc>ui_main.action_stop->setIcon(QIcon::fromTheme("media-playback-stop"));
            #pxc>ui_main.action_play->setIcon(QIcon::fromTheme("media-playback-start"));
            #pxc>ui_main.action_next->setIcon(QIcon::fromTheme("media-skip-forward"));
            #pxc>ui_main.action_forward->setIcon(QIcon::fromTheme("media-seek-forward"));
            #pxc>ui_main.action_mark_set->setIcon(QIcon::fromTheme("list-add"));
            #pxc>ui_main.action_mark_remove->setIcon(QIcon::fromTheme("list-remove"));
            #pxc>ui_main.action_selectmodel->setIcon(icon3);
            #pxc>ui_main.action_selectmodel_back->setIcon(QIcon::fromTheme("back"));
            #pxc>ui_main.action_preferences->setIcon(QIcon::fromTheme("document-properties"));
            #pxc>ui_main.action_info->setIcon(QIcon::fromTheme("help-about"));
            #pxc>ui_main.action_help->setIcon(QIcon::fromTheme("help"));
            
            #pxc>// widgets that are not created with Qt Designer: toolbar_play
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_selectmodel);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_new_solved);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_challenge);
            #pxc>ui_main.toolbar_play->addSeparator();
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_rewind);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_previous);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_stop);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_play);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_next);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_forward);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_mark_set);
            #pxc>ui_main.toolbar_play->addAction(ui_main.action_mark_remove);
            
            #pxc>// …: toolbar_selectmodel
            #pxc>ui_main.toolbar_selectmodel->addAction(ui_main.action_selectmodel_back);
            #pxc>label_selectmodel = new QLabel();
            #pxc>ui_main.toolbar_selectmodel->addWidget(label_selectmodel);
            #pxc>ui_main.toolbar_selectmodel->setVisible(false);
            
            #pxc>// …: toolbar_common
            #pxc>QWidget *spacer = new QWidget();
            #pxc>spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
            #pxc>ui_main.toolbar_common->addWidget(spacer);
            
            #pxc>// …: menu button
            #pxc>QToolButton *menubutton = new QToolButton(window);
            #pxc>menubutton->setIcon(QIcon::fromTheme("format-justify-fill")); //XXX: hmm
            #pxc>menubutton->setPopupMode(QToolButton::InstantPopup);
            #pxc>QMenu *menu = new QMenu(menubutton);
            #pxc>menu->addAction(ui_main.action_initial_state);
            #pxc>menu->addAction(ui_main.action_reset_rotation);
            #pxc>menu->addSeparator();
            #pxc>menu->addAction(ui_main.action_editbar);
            #pxc>menu->addAction(ui_main.action_statusbar);
            #pxc>menu->addSeparator();
            #pxc>menu->addAction(ui_main.action_preferences);
            #pxc>menu->addAction(ui_main.action_help);
            #pxc>menu->addAction(ui_main.action_info);
            #pxc>menu->addSeparator();
            #pxc>menu->addAction(ui_main.action_quit);
            #pxc>menubutton->setMenu(menu);
            #pxc>ui_main.toolbar_common->addWidget(menubutton);
            
            #pxc>// add action to main window for shortcuts
            #pxc>window->addAction(ui_main.action_challenge);
            #pxc>window->addAction(ui_main.action_new_solved);
            #pxc>window->addAction(ui_main.action_quit);
            #pxc>window->addAction(ui_main.action_selectmodel);
            #pxc>window->addAction(ui_main.action_initial_state);
            #pxc>window->addAction(ui_main.action_reset_rotation);
            #pxc>window->addAction(ui_main.action_preferences);
            #pxc>window->addAction(ui_main.action_statusbar);
            #pxc>window->addAction(ui_main.action_info);
            #pxc>window->addAction(ui_main.action_rewind);
            #pxc>window->addAction(ui_main.action_rewind);
            #pxc>window->addAction(ui_main.action_previous);
            #pxc>window->addAction(ui_main.action_previous);
            #pxc>window->addAction(ui_main.action_stop);
            #pxc>window->addAction(ui_main.action_stop);
            #pxc>window->addAction(ui_main.action_play);
            #pxc>window->addAction(ui_main.action_play);
            #pxc>window->addAction(ui_main.action_next);
            #pxc>window->addAction(ui_main.action_next);
            #pxc>window->addAction(ui_main.action_forward);
            #pxc>window->addAction(ui_main.action_forward);
            #pxc>window->addAction(ui_main.action_mark_set);
            #pxc>window->addAction(ui_main.action_mark_set);
            #pxc>window->addAction(ui_main.action_mark_remove);
            #pxc>window->addAction(ui_main.action_mark_remove);
            #pxc>window->addAction(ui_main.action_editbar);
            #pxc>window->addAction(ui_main.action_help);
            #pxc>window->addAction(ui_main.action_help);
            
            #pxc>// rest
            #pxc>disable_tooltips();
            #pxc>active_plugin_group = 0;
        #pxc>}
    #pxm-FUNC CHP
    def disable_tooltips():  pass
        #pxc>{
            #pxc>//pita: If a action has no tooltip, the appropriate toolbutton uses the actions text for the tooltip.
            #pxc>//      To remove tooltips you have to remove the tooltips for the appropriate toolbutton,
            #pxc>//      but every time you change something on the action, the button will recreate the tooltip
            #pxc>ui_main.toolbar_play->widgetForAction(ui_main.action_rewind)->setToolTip(QStringLiteral(""));
            #pxc>ui_main.toolbar_play->widgetForAction(ui_main.action_previous)->setToolTip(QStringLiteral(""));
            #pxc>ui_main.toolbar_play->widgetForAction(ui_main.action_stop)->setToolTip(QStringLiteral(""));
            #pxc>ui_main.toolbar_play->widgetForAction(ui_main.action_play)->setToolTip(QStringLiteral(""));
            #pxc>ui_main.toolbar_play->widgetForAction(ui_main.action_next)->setToolTip(QStringLiteral(""));
            #pxc>ui_main.toolbar_play->widgetForAction(ui_main.action_forward)->setToolTip(QStringLiteral(""));
        #pxc>}
    #pxm-FUNC CHP
    def set_shortcuts(key1:1, key2:1, key3:1, key4:'const QKeySequence &'):  pass
        #pxc>{
            #pxc>ui_main.action_selectmodel->setShortcut(key1);
            #pxc>ui_main.action_initial_state->setShortcut(key2);
            #pxc>ui_main.action_reset_rotation->setShortcut(key3);
            #pxc>ui_main.action_preferences->setShortcut(key4);
        #pxc>}
    #pxm-FUNC CHP
    def add_widgets(window:'QMainWindow *', edit:'QLineEdit *', drawing:'QOpenGLWidget *'):  pass
        #pxc>{
            #pxc>edit->setObjectName("edit_moves");
            #pxc>edit->setFrame(false);
            #pxc>ui_main.layout_moves->insertWidget(1, edit);
            #pxc>drawing->setObjectName("drawingarea");
            #pxc>ui_main.verticalLayout->addWidget(drawing);
            #pxc>window->setTabOrder(edit, drawing);
            #pxc>window->setTabOrder(drawing, ui_main.box_sidepane);
            #pxc>drawing->setFocus(Qt::OtherFocusReason);
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def set_toolbar_visible(visible:bool):  pass
        #pxc>{ ui_main.toolbar_play->setVisible(visible); }
    #pxm-FUNC CHP nogil
    def set_frame_editbar_visible(visible:bool):  pass
        #pxc>{ ui_main.frame_editbar->setVisible(visible); }
    #pxm-FUNC CHP nogil
    def set_statusbar_visible(visible:bool):  pass
        #pxc>{ ui_main.statusbar->setVisible(visible); }
    #pxm-FUNC CHP nogil
    def hide_controls(): pass
        #pxc>{
            #pxc>ui_main.toolbar_play->setVisible(false);
            #pxc>ui_main.toolbar_common->setVisible(false);
            #pxc>ui_main.frame_editbar->setVisible(false);
            #pxc>ui_main.statusbar->setVisible(false);
            #pxc>ui_main.box_sidepane->setVisible(false);
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def set_visible(debug_text_visible:bool, editbar_visible:bool, statusbar_visible:bool):  pass
        #pxc>{
            #pxc>ui_main.label_debug_text->setVisible(debug_text_visible);
            #pxc>ui_main.action_editbar->setChecked(editbar_visible);
            #pxc>ui_main.action_statusbar->setChecked(statusbar_visible);
            #pxc>ui_main.frame_editbar->setVisible(editbar_visible);
            #pxc>ui_main.statusbar->setVisible(statusbar_visible);
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def get_splitter_sizes(s1:'int &', s2:'int &'):  pass
        #pxc>{
            #pxc>QList<int> list(ui_main.splitter->sizes());
            #pxc>s1 = list.at(0);
            #pxc>s2 = list.at(1);
        #pxc>}
    #pxm-FUNC CHP nogil
    def set_splitter_sizes(s1:int, s2:int):  pass
        #pxc>{
            #pxc>QList<int> list;
            #pxc>list.append(s1);
            #pxc>list.append(s2);
            #pxc>ui_main.splitter->setSizes(list);
        #pxc>}
    #pxm-FUNC CHP nogil
    def splitter_update_minimumsize():  pass
        #pxc>{
            #pxc>int w0, w1;
            #pxc>get_splitter_sizes(w0, w1);
            #pxc>QSize s0 = ui_main.splitter->widget(0)->minimumSizeHint();
            #pxc>QSize s1 = ui_main.splitter->widget(1)->minimumSizeHint();
            #pxc>int w = s0.width() + ui_main.splitter->handleWidth();
            #pxc>int h = std::max(s0.height(), s1.height());
            #pxc>w += w1==0 ? 6 : s1.width();
            #pxc>ui_main.splitter->setMinimumSize(w, h);
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def set_debug_text(text:str):  pass
        #pxc>{ ui_main.label_debug_text->setText(text); }
        
    #pxm-FUNC CHP nogil
    def set_toolbar_state(a:int, b:int, c:int, d:int, e:int):  pass
        #pxc>{
            #pxc>ui_main.action_rewind->setEnabled(a);
            #pxc>ui_main.action_previous->setEnabled(a);
            #pxc>ui_main.action_play->setEnabled(b);
            #pxc>ui_main.action_next->setEnabled(b);
            #pxc>ui_main.action_forward->setEnabled(b);
            #pxc>ui_main.action_mark_set->setEnabled(c);
            #pxc>ui_main.action_mark_remove->setEnabled(c);
            #pxc>ui_main.action_stop->setVisible(d);
            #pxc>ui_main.action_play->setVisible(!d);
            #pxc>ui_main.action_mark_set->setVisible(e);
            #pxc>ui_main.action_mark_remove->setVisible(!e);
            #pxc>disable_tooltips();
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def clear_sidepane():  pass
        #pxc>{
            #pxc>QLayoutItem *child;
            #pxc>while ((child = ui_main.layout_sidepane->takeAt(0)) != 0) {
                #pxc>delete child;
            #pxc>}
        #pxc>}
    #pxm-FUNC CHP nogil
    def create_sidepane_button(text:str)->'QPushButton *':  pass
        #pxc>{
            #pxc>QPushButton *button = new QPushButton(ui_main.box_sidepane);
            #pxc>button->setText(text);
            #pxc>button->setFlat(true);
            #pxc>button->setFocusPolicy(Qt::TabFocus);
            #pxc>ui_main.layout_sidepane->addWidget(button);
            #pxc>return button;
        #pxc>}
    #pxm-FUNC CHP nogil
    def create_sidepane_treeview(treestore:'QStandardItemModel *', i:int)->'QTreeView *':  pass
        #pxc>{
            #pxc>QTreeView *treeview = new QTreeView(ui_main.box_sidepane);
            #pxc>treeview->setFrameShape(QFrame::NoFrame);
            #pxc>treeview->setEditTriggers(QAbstractItemView::NoEditTriggers);
            #pxc>treeview->setUniformRowHeights(true);
            #pxc>treeview->setAnimated(true);
            #pxc>treeview->setHeaderHidden(true);
            #pxc>treeview->hide();
            #pxc>treeview->setModel(treestore);
            #pxc>treeview->setRootIndex(treestore->index(i, 0));
            #pxc>ui_main.layout_sidepane->addWidget(treeview);
            #pxc>return treeview;
        #pxc>}
    #pxm-FUNC CHP nogil
    def set_active_plugin_group(i:int):  pass
        #pxc>{
            #pxc>QLayoutItem *item_treeview;
            #pxc>if (i == -1)  i = active_plugin_group;
            #pxc>item_treeview = ui_main.layout_sidepane->itemAt(active_plugin_group*2+1);
            #pxc>if (item_treeview != NULL)
                #pxc>item_treeview->widget()->hide();
            #pxc>item_treeview = ui_main.layout_sidepane->itemAt(i*2+1);
            #pxc>if (item_treeview != NULL)
                #pxc>item_treeview->widget()->show();
            #pxc>active_plugin_group = i;
        #pxc>}
    #pxm-FUNC CHP nogil
    def set_active_plugin_group_by_obj(obj:'QObject *'):  pass
        #pxc>{
            #pxc>int i = ui_main.layout_sidepane->indexOf((QWidget*)obj);
            #pxc>this->set_active_plugin_group(i/2);
        #pxc>}
    #pxm-FUNC CHP nogil
    def get_plugin_group_count()->int:  pass
        #pxc>{ return ui_main.layout_sidepane->count() / 2; }
    #pxm-FUNC CHP nogil
    def root_index(i:int)->'QModelIndex':  pass
        #pxc>{ return ((QTreeView*)ui_main.layout_sidepane->itemAt(i*2+1)->widget())->rootIndex(); }
    #pxm-FUNC CHP nogil
    def hide_row(i:int, hide:bool):  pass
        #pxc>{
            #pxc>QWidget *button = ui_main.layout_sidepane->itemAt(i*2)->widget();
            #pxc>if (hide) {
            #pxc>    button->hide();
            #pxc>    if (i == active_plugin_group)  set_active_plugin_group(0);
            #pxc>} else {
            #pxc>    button->show();
            #pxc>}
        #pxc>}
    #pxm-FUNC CHP nogil
    def set_row_hidden(i:int, row:int, index:'const QModelIndex &', hide:bool):  pass
        #pxc>{ ((QTreeView*)ui_main.layout_sidepane->itemAt(i*2+1)->widget())->setRowHidden(row, index, hide); }
    #pxm-FUNC CHP nogil
    def splitter_pos()->int:  pass
        #pxc>{ return ui_main.splitter->sizes()[0]; }
        
    #pxh>// ####
    #pxm-FUNC CHP nogil
    def setupUi_pref(window:'QDialog *', speed:int, speed_min:int, speed_max:int, mirror_min:int, mirror_max:int, _movekeys:'QStandardItemModel *'):  pass
        #pxc>{
            #pxc>liststore_movekeys = _movekeys;
            #pxc>liststore_movekeys->setObjectName("liststore_movekeys");
            #pxc>liststore_faces = new QStandardItemModel(window);
            #pxc>liststore_faces->setObjectName("liststore_faces");
            #pxc>
            #pxc>ui_pref.setupUi(window);
            #pxc>
            #pxc>ui_pref.button_animspeed_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_pref.button_shader_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_pref.button_antialiasing_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_pref.button_mirror_faces_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_pref.button_movekey_add->setIcon(QIcon::fromTheme("list-add"));
            #pxc>ui_pref.button_movekey_remove->setIcon(QIcon::fromTheme("list-remove"));
            #pxc>ui_pref.button_movekey_reset->setIcon(QIcon::fromTheme("document-revert"));
            #pxc>ui_pref.button_color_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_pref.button_image_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>ui_pref.button_background_color_reset->setIcon(QIcon::fromTheme("edit-clear"));
            #pxc>
            #pxc>ui_pref.buttonBox->button(QDialogButtonBox::Close)->setDefault(true);
            #pxc>ui_pref.label_needs_restarted->setVisible(false);
            #pxc>QPalette pal(ui_pref.label_needs_restarted->palette());
            #pxc>pal.setColor(QPalette::WindowText, QColor(170, 0, 0));
            #pxc>ui_pref.label_needs_restarted->setPalette(pal);
            #pxc>
            #pxc>// graphic tab
            #pxc>ui_pref.slider_animspeed->setRange(speed_min, speed_max);
            #pxc>ui_pref.slider_animspeed->setValue(speed);
            #pxc>ui_pref.spinbox_mirror_faces->setRange(mirror_min, mirror_max);
            #pxc>
            #pxc>// keys tab
            #pxc>liststore_movekeys->setColumnCount(2);
            #pxc>ui_pref.listview_movekeys->setModel(liststore_movekeys);
            #pxc>ShortcutDelegate *shortcut_delegate = new ShortcutDelegate(ui_pref.listview_movekeys);
            #pxc>ui_pref.listview_movekeys->setItemDelegateForColumn(1, shortcut_delegate);
            #pxc>
            #pxc>// theme tab
            #pxc>liststore_faces_width = 30;
            #pxc>iconengine_color = new ColorIconEngine();
            #pxc>ui_pref.button_color->setIcon(QIcon(iconengine_color));
            #pxc>int height = ui_pref.button_color->iconSize().height();
            #pxc>ui_pref.button_color->setIconSize(
            #pxc>        QSize(height * ui_pref.button_color->width() / ui_pref.button_color->height(), height));
            #pxc>ui_pref.button_color->setText(QStringLiteral(""));
            #pxc>iconengine_background_color = new ColorIconEngine();
            #pxc>ui_pref.button_background_color->setIcon(QIcon(iconengine_background_color));
            #pxc>height = ui_pref.button_background_color->iconSize().height();
            #pxc>ui_pref.button_background_color->setIconSize(
            #pxc>        QSize(height * ui_pref.button_background_color->width() / ui_pref.button_background_color->height(), height));
            #pxc>ui_pref.button_background_color->setText(QStringLiteral(""));
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def preferences_block_signals(block:bool):  pass
        #pxc>{
            #pxc>ui_pref.combobox_shader->blockSignals(block);
            #pxc>ui_pref.combobox_samples->blockSignals(block);
            #pxc>liststore_movekeys->blockSignals(block);
        #pxc>}
    #pxm-FUNC CHP nogil
    def combobox_add_shaderitem(name:str, nick:str):  pass
        #pxc>{ ui_pref.combobox_shader->addItem(name, nick); }
    #pxm-FUNC CHP nogil
    def combobox_add_samplesitem(name:str, nick:str):  pass
        #pxc>{ ui_pref.combobox_samples->addItem(name, nick); }
    #pxm-FUNC CHP nogil
    def add_movekey_row():  pass
        #pxc>{
            #pxc>int row = ui_pref.listview_movekeys->currentIndex().row();
            #pxc>liststore_movekeys->insertRow(row);
            #pxc>liststore_movekeys->setItem(row, 0, new QStandardItem(""));
            #pxc>liststore_movekeys->setItem(row, 1, new QStandardItem(""));
            #pxc>QModelIndex index = liststore_movekeys->index(row, 0);
            #pxc>ui_pref.listview_movekeys->setCurrentIndex(index);
            #pxc>ui_pref.listview_movekeys->edit(index);
        #pxc>}
    #pxm-FUNC CHP nogil
    def remove_movekey_row():  pass
        #pxc>{
            #pxc>int row = ui_pref.listview_movekeys->currentIndex().row();
            #pxc>liststore_movekeys->takeRow(row);
        #pxc>}
    #pxm-FUNC CHP nogil
    def add_liststore_faces_row(name:str, key:str):  pass
        #pxc>{
            #pxc>QStandardItem *item = new QStandardItem(name);
            #pxc>item->setData(QVariant(key));
            #pxc>liststore_faces->appendRow(item);
            #pxc>//XXX: workaround, listview_faces should automatically set to the correct width
            #pxc>QFontMetrics fm(ui_pref.listview_faces->font());
            #pxc>int width = fm.width(name) + 20;  // 20 for the scroll bar
            #pxc>if (liststore_faces_width < width)  liststore_faces_width = width;
        #pxc>}
    #pxm-FUNC CHP nogil
    def finalize_liststore_faces(window:'QWidget *'):  pass
        #pxc>{
            #pxc>ui_pref.listview_faces->setMaximumWidth(liststore_faces_width);
            #pxc>ui_pref.listview_faces->setModel(liststore_faces);
            #pxc>QObject::connect(ui_pref.listview_faces->selectionModel(),
            #pxc>                   SIGNAL(currentRowChanged(const QModelIndex &, const QModelIndex &)),
            #pxc>                 window,
            #pxc>                   SLOT(_on_listview_faces_currentRowChanged(const QModelIndex &)));
            #pxc>ui_pref.listview_faces->setCurrentIndex(liststore_faces->index(0, 0));
        #pxc>}
    #pxm-FUNC CHP nogil
    def get_liststore_faces_facekey(index:'const QModelIndex &')->str:  pass
        #pxc>{ return liststore_faces->itemFromIndex(index)->data().toString(); }
    #pxm-FUNC CHP nogil
    def set_button_color(color:str):  pass
        #pxc>{
            #pxc>iconengine_color->color = color;
            #pxc>ui_pref.button_color->update();
        #pxc>}
    #pxm-FUNC CHP nogil
    def color_dialog(parent:'QWidget *', color:str)->str:  pass
        #pxc>{
            #pxc>QColorDialog dialog(parent);
            #pxc>dialog.setOption(QColorDialog::DontUseNativeDialog);
            #pxc>dialog.setCurrentColor(QColor(color));
            #pxc>if (dialog.exec() == QDialog::Accepted)
                #pxc>return dialog.currentColor().name();
            #pxc>return QString();
        #pxc>}
    #pxm-FUNC CHP nogil
    def set_imagemode(imagemode:int):  pass
        #pxc>{
            #pxc>if (imagemode == 0)
                #pxc>ui_pref.radiobutton_tiled->setChecked(true);
            #pxc>else if (imagemode == 1)
                #pxc>ui_pref.radiobutton_mosaic->setChecked(true);
        #pxc>}
    #pxm-FUNC CHP nogil
    def add_combobox_image_item(icon:'const QIcon &', text:str, filename:str):  pass
        #pxc>{ ui_pref.combobox_image->addItem(icon, text, filename); }
    #pxm-FUNC CHP nogil
    def set_combobox_image(index_icon:int, icon:'const QIcon &', index:int):  pass
        #pxc>{
            #pxc>ui_pref.combobox_image->setItemIcon(index_icon, icon);
            #pxc>if (index >= 0)  ui_pref.combobox_image->setCurrentIndex(index);
        #pxc>}
    #pxh>
    #pxm-FUNC CHP nogil
    def set_button_background_color(color:str):  pass
        #pxc>{
            #pxc>iconengine_background_color->color = color;
            #pxc>ui_pref.button_background_color->update();
        #pxc>}
    
    #pxm-FUNC CHP nogil
    def fill_page(items:'QVariant', title:str):  pass
        #pxc>{
            #pxc>ui_main.listwidget->clear();
            #pxc>QList<QObject*> list = items.value<QList<QObject*> >();
            #pxc>for (int i=0; i < list.size(); ++i) {
                #pxc>QObject *obj = list.at(i);
                #pxc>QString text = obj->property("text").value<QString>();
                #pxc>QString filename = obj->property("key").value<QString>();
                #pxc>QListWidgetItem *item = new QListWidgetItem(QIcon(filename), text);
                #pxc>item->setTextAlignment(Qt::AlignHCenter | Qt::AlignBottom);
                #pxc>ui_main.listwidget->addItem(item);
            #pxc>}
            #pxc>label_selectmodel->setText(title);
        #pxc>}
    #pxm-FUNC CHP nogil
    def set_page(i:int):  pass
        #pxc>{
            #pxc>ui_main.stackedwidget->setCurrentIndex(i);
            #pxc>ui_main.toolbar_play->setVisible(i==0);
            #pxc>ui_main.toolbar_selectmodel->setVisible(i!=0);
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def setupUi_help(window:'QDialog *', helptext:str):  pass
        #pxc>{
            #pxc>ui_help.setupUi(window);
            #pxc>ui_help.text_help->setHtml(helptext);
        #pxc>}
        
    #pxm-FUNC CHP nogil
    def setupUi_about(window:'QDialog *'):  pass
        #pxc>{
            #pxc>ui_about.setupUi(window);
            #pxc>index_tab_about = ui_about.tab_widget->indexOf(ui_about.tab_about);
            #pxc>index_tab_license = ui_about.tab_widget->indexOf(ui_about.tab_license);
            #pxc>// About tab animation
            #pxc>scrollbar = ui_about.text_translators->verticalScrollBar();
            #pxc>animation.setTargetObject(scrollbar);
            #pxc>animation.setPropertyName(QByteArray("value"));
            #pxc>animation.setLoopCount(-1);
            #pxc>ui_about.text_translators->viewport()->installEventFilter(window);
        #pxc>}
    #pxm-FUNC CHP nogil
    def fill_header(fileName:str, appname:str, version:str, desc:str):  pass
        #pxc>{
            #pxc>ui_about.label_icon->setPixmap(QPixmap(fileName));
            #pxc>ui_about.label_appname->setText(appname);
            #pxc>ui_about.label_version->setText(version);
            #pxc>ui_about.label_description->setText(desc);
        #pxc>}
    #pxm-FUNC CHP nogil
    def fill_about_tab(copyr:str, website:str, translators:str):  pass
        #pxc>{
            #pxc>ui_about.label_copyright->setText(copyr);
            #pxc>ui_about.label_website->setText(website);
            #pxc>ui_about.text_translators->setHtml(translators);
        #pxc>}
    #pxm-FUNC CHP nogil
    def fill_contribute_tab(contribute:str):  pass
        #pxc>{ ui_about.label_contribute->setText(contribute); }
    #pxm-FUNC CHP nogil
    def fill_license_tab(license_short:str, license_notfound:bool, license_full:str):  pass
        #pxc>{
            #pxc>ui_about.text_license_short->hide();
            #pxc>ui_about.text_license_full->hide();
            #pxc>ui_about.text_license_short->setHtml(license_short);
            #pxc>if (license_notfound)  ui_about.text_license_full->setLineWrapMode(QTextEdit::WidgetWidth);
            #pxc>ui_about.text_license_full->setHtml(license_full);
        #pxc>}
    #pxm-FUNC CHP nogil
    def tab_widget_currentChanged(index:int):  pass
        #pxc>{
            #pxc>if (index == index_tab_about) {
                #pxc>animation.resume();
                #pxc>scrollbar->hide();
            #pxc>} else {
                #pxc>animation.pause();
                #pxc>scrollbar->show();
            #pxc>}
            #pxc>if (index == index_tab_license) {
                #pxc>ui_about.text_license_short->setVisible(true);
                #pxc>ui_about.text_license_full->setVisible(false);
            #pxc>}
        #pxc>}
    #pxm-FUNC CHP nogil
    def update_animation(first:bool):  pass
        #pxc>{
            #pxc>int smin, smax;
            #pxc>if (first and animation.state() != QAbstractAnimation::Stopped) return;
            #pxc>smin = scrollbar->minimum();
            #pxc>smax = scrollbar->maximum();
            #pxc>if (smax > smin) {
                #pxc>animation.setDuration((smax-smin) * 40);
                #pxc>animation.setKeyValueAt(0., smin);
                #pxc>animation.setKeyValueAt(0.04, smin);
                #pxc>animation.setKeyValueAt(0.50, smax);
                #pxc>animation.setKeyValueAt(0.54, smax);
                #pxc>animation.setKeyValueAt(1., smin);
            #pxc>}
            #pxc>if (first) {
                #pxc>scrollbar->hide();
                #pxc>animation.start();
            #pxc>}
        #pxc>}
    #pxm-FUNC CHP nogil
    def event_filter_stop_animation(event:'QEvent *'):  pass
        #pxc>{
            #pxc>//assert watched == ui_about.text_translators.viewport()
            #pxc>if ((event->type() == QEvent::MouseButtonPress) || (event->type() == QEvent::Wheel)) {
                #pxc>animation.pause();
                #pxc>scrollbar->show();
            #pxc>}
        #pxc>}
    #pxm-FUNC CHP nogil
    def show_full_license(local:bool, url:'const QUrl &'):  pass
        #pxc>{
            #pxc>if (local) {
                #pxc>ui_about.text_license_short->setVisible(false);
                #pxc>ui_about.text_license_full->setVisible(true);
            #pxc>} else
                #pxc>QDesktopServices::openUrl(url);
        #pxc>}

#pxm>CPPCLASS_END
    

#pxm-FUNC CHP
def mk_qstr(pstr:'const char *', length:int)->str:  return pstr
    #pxc>{ return QString::fromUtf8(pstr, length); }
    
#pxm>DEDENT

#px/cdef UIContainer qtui
qtui = UIContainer()

#pxm-FUNC PD
def init_gettext(gettext):
    global _
    _ = gettext
#px/cdef QString gettext_translate(const char *text, const char *disambiguation) with gil:
def gettext_translate(text, disambiguation):
    pytext = _(text.decode('utf-8')).encode('utf-8')
    return mk_qstr(pytext, len(pytext))
    
    
#XXX: Cython cannot export C functions, so this wrapper functions are needed.
#TODO: Makro for generating wrapper functions or exported function pointers

#pxm-FUNC PD
def setupUi_main(window:'QMainWindow *', icon1:1, icon2:1, icon3:'const QIcon &', key1:1, key2:'const QKeySequence &', editbar_visible:bool, statusbar_visible:bool):
    qtui.setupUi_main(window, icon1, icon2, icon3, key1, key2)
    qtui.set_visible(debug & (DEBUG_DRAW | DEBUG_FPS), editbar_visible, statusbar_visible)
    
#pxm-FUNC PD
def setupUi_pref(window:'QDialog *', speed:int, speed_min:int, speed_max:int, mirror_min:int, mirror_max:int, liststore_movekeys:'QStandardItemModel *'):
    qtui.setupUi_pref(window, speed, speed_min, speed_max, mirror_min, mirror_max, liststore_movekeys)
    
#pxm-FUNC PD
def setupUi_help(window:'QDialog *', helptext:str):
    qtui.setupUi_help(window, helptext)
    
#pxm-FUNC PD
def setupUi_about(window:'QDialog *'):
    qtui.setupUi_about(window)
    
#pxm-FUNC PD
def set_shortcuts(key1:1, key2:1, key3:1, key4:'const QKeySequence &'):
    qtui.set_shortcuts(key1, key2, key3, key4)
#pxm-FUNC PD
def add_widgets(window:'QMainWindow *', edit:'QLineEdit *', drawing:'QOpenGLWidget *'):
    qtui.add_widgets(window, edit, drawing)
    
#pxm-FUNC PD nogil
def set_toolbar_visible(visible:bool):
    qtui.set_toolbar_visible(visible)
#pxm-FUNC PD nogil
def set_frame_editbar_visible(visible:bool):
    qtui.set_frame_editbar_visible(visible)
#pxm-FUNC PD nogil
def set_statusbar_visible(visible:bool):
    qtui.set_statusbar_visible(visible)
#pxm-FUNC PD nogil
def hide_controls():
    qtui.hide_controls()
#pxm-FUNC PD nogil
def get_splitter_sizes(s1:'int &', s2:'int &'):
    qtui.get_splitter_sizes(s1, s2)
#pxm-FUNC PD nogil
def set_splitter_sizes(s1:int, s2:int):
    qtui.set_splitter_sizes(s1, s2)
#pxm-FUNC PD nogil
def splitter_update_minimumsize():
    qtui.splitter_update_minimumsize()
#pxm-FUNC PD nogil
def set_debug_text(text:str):
    qtui.set_debug_text(text)
    
#pxm-FUNC PD nogil
def set_toolbar_state(a:int, b:int, c:int, d:int, e:int):
    qtui.set_toolbar_state(a, b, c, d, e)
    
#pxm-FUNC PD nogil
def clear_sidepane():
    qtui.clear_sidepane()
#pxm-FUNC PD nogil
def create_sidepane_button(text:str)->'QPushButton *':
    return qtui.create_sidepane_button(text)
#pxm-FUNC PD nogil
def create_sidepane_treeview(treestore:'QStandardItemModel *', i:int)->'QTreeView *':
    return qtui.create_sidepane_treeview(treestore, i)
#pxm-FUNC PD nogil
def set_active_plugin_group(i:int):
    qtui.set_active_plugin_group(i)
#pxm-FUNC PD nogil
def set_active_plugin_group_by_obj(obj:'QObject *'):
    qtui.set_active_plugin_group_by_obj(obj)
#pxm-FUNC PD nogil
def get_plugin_group_count()->int:
    return qtui.get_plugin_group_count()
#pxm-FUNC PD nogil
def root_index(i:int)->'QModelIndex':
    return qtui.root_index(i)
#pxm-FUNC PD nogil
def hide_row(i:int, hide:bool):
    qtui.hide_row(i, hide)
#pxm-FUNC PD nogil
def set_row_hidden(treeview:int, row:int, index:'const QModelIndex &', hide:bool):
    qtui.set_row_hidden(treeview, row, index, hide)
#pxm-FUNC PD nogil
def splitter_pos()->int:
    return qtui.splitter_pos()
    
#pxm-FUNC PD nogil
def preferences_block_signals(block:bool):
    qtui.preferences_block_signals(block)
#pxm-FUNC PD nogil
def combobox_add_shaderitem(name:str, nick:str):
    qtui.combobox_add_shaderitem(name, nick)
#pxm-FUNC PD nogil
def combobox_add_samplesitem(name:str, nick:str):
    qtui.combobox_add_samplesitem(name, nick)
#pxm-FUNC PD nogil
def add_movekey_row():
    qtui.add_movekey_row()
#pxm-FUNC PD nogil
def remove_movekey_row():
    qtui.remove_movekey_row()
#pxm-FUNC PD nogil
def add_liststore_faces_row(name:str, key:str):
    qtui.add_liststore_faces_row(name, key)
#pxm-FUNC PD nogil
def finalize_liststore_faces(window:'QWidget *'):
    qtui.finalize_liststore_faces(window)
#pxm-FUNC PD nogil
def get_liststore_faces_facekey(index:'const QModelIndex &')->str:
    return qtui.get_liststore_faces_facekey(index)
#pxm-FUNC PD nogil
def set_button_color(color:str):
    qtui.set_button_color(color)
#pxm-FUNC PD nogil
def color_dialog(parent:'QWidget *', color:str)->str:
    return qtui.color_dialog(parent, color)
#pxm-FUNC PD nogil
def set_imagemode(imagemode:int):
    qtui.set_imagemode(imagemode)
#pxm-FUNC PD nogil
def add_combobox_image_item(icon:'const QIcon &', text:str, filename:str):
    qtui.add_combobox_image_item(icon, text, filename)
#pxm-FUNC PD nogil
def set_combobox_image(index_icon:int, icon:'const QIcon &', index:int):
    qtui.set_combobox_image(index_icon, icon, index)
#pxm-FUNC PD nogil
def set_button_background_color(color:str):
    qtui.set_button_background_color(color)
    
#pxm-FUNC PD nogil
def fill_page(items:'QVariant', title:str):
    qtui.fill_page(items, title)
#pxm-FUNC PD nogil
def set_page(i:int):
    qtui.set_page(i)
    
#pxm-FUNC PD nogil
def fill_header(fileName:str, appname:str, version:str, desc:str):
    qtui.fill_header(fileName, appname, version, desc)
#pxm-FUNC PD nogil
def fill_about_tab(copyr:str, website:str, translators:str):
    qtui.fill_about_tab(copyr, website, translators)
#pxm-FUNC PD nogil
def fill_contribute_tab(contribute:str):
    qtui.fill_contribute_tab(contribute)
#pxm-FUNC PD nogil
def fill_license_tab(license_short:str, license_notfound:bool, license_full:str):
    qtui.fill_license_tab(license_short, license_notfound, license_full)
#pxm-FUNC PD nogil
def tab_widget_currentChanged(index:int):
    qtui.tab_widget_currentChanged(index)
#pxm-FUNC PD nogil
def update_animation(first:bool):
    qtui.update_animation(first)
#pxm-FUNC PD nogil
def event_filter_stop_animation(event:'QEvent *'):
    qtui.event_filter_stop_animation(event)
#pxm-FUNC PD nogil
def show_full_license(local:bool, url:'const QUrl &'):
    qtui.show_full_license(local, url)
    




# -*- coding: utf-8 -*-

#  Copyright © 2013-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .debug import DEBUG_SHADER

N_ = lambda s: s

if DEBUG_SHADER:
    import os
    from . import config
    def shaders(names):
        for filename in os.listdir(config.SHADER_DIR):
            name = os.path.splitext(filename)[0]
            if name not in names:
                names.append(name)
        return names
else:
    shaders = lambda s: s
    
def tuple_validator(minlen, maxlen, itemtype, value, valuerange=None):
    if type(value) is not tuple:
        return False
    if not (minlen <= len(value) <= maxlen):
        return False
    for v in value:
        if type(v) is not itemtype:
            return False
        if valuerange is not None and not valuerange[0] <= v <= valuerange[1]:
            return False
    return True
    
def validate_mtype(v):
    from .model import Model
    return v in Model.cache_index['types']
    
def validate_draw_accels(v):
    if type(v) is not list: return False
    for t in v:
        if type(t) is not tuple or len(t) != 2: return False
        if type(t[0]) is not str or type(t[1]) is not str: return False
    return True
    
deprecated = type('DeprecatedType', (), {})
    
def migrate_game_size_blocks_moves_position(settings):
    mtype = settings['game.type']
    size = settings['game.size']
    blocks = settings['game.blocks']
    moves = settings['game.moves']
    position = settings['game.position']
    
    if not tuple_validator(3, 100, int, size, (1, 10)):
        return
    from .model import Model
    defaultsize = Model.cache_index['type'][mtype]['defaultsize']
    if len(size) < len(defaultsize):
        return
    if len(size) > len(defaultsize):
        size = size[:len(defaultsize)]
    try:
        size = Model.cache_index['normsize'][mtype][size]
    except KeyError:
        return
    settings['games',mtype,'size'] = size
    
    if type(blocks) is not str:
        return
    
    if type(moves) is not str:
        moves = ''
    
    if type(position) is not int or position < 0:
        position = 0
    return mtype, size, blocks, moves, position
    
def migrate_theme_face_N_color_image_mode(settings):
    mtype = settings['game.type']
    facenames = ('Up', 'Down', 'Left', 'Right', 'Front', 'Back')
    for i, facename in enumerate(facenames):
        color = settings['theme.face',i,'color']
        image = settings['theme.face',i,'image']
        mode = settings['theme.face',i,'mode']
        if type(color) is str:  settings['theme.faces',facename,'color'] = color
        if type(image) is str:  settings['theme.faces',facename,'image'] = image
        if type(mode) in ['tiled', 'mosaic']:  settings['theme.faces',facename,'mode'] = mode
        
def migrate_2_3(settings):
    for facename in ('Up', 'Down', 'Left', 'Right', 'Front', 'Back',
                     'Front_Left', 'Front_Right', 'Back_Right', 'Back_Left'):
        for attr in ('color', 'image', 'mode'):
            oldkey = '.'.join(('theme.faces', facename, attr))
            if oldkey in settings.keystore.keystore:
                settings['theme.faces', facename.lower(), attr] = settings[oldkey]
                
    
schema = {
    # key:              (default,   range/enum/validator)
    #                               None: value without restriction
    #                               tuple: contains two values (min, max)
    #                               list: contains strings for the enum text,
    #                                     the index is the enum value
    #                               function: returns True, if value is valid
    'version':              (1,         lambda v: type(v) is int),
    'window.size':          ((850, 650),lambda v: tuple_validator(2, 2, int, v, (10,10000))),
    'window.divider':       (620,       (0, 10000)),
    'window.sidepane_width':(220,       (0, 10000)),
    'window.toolbar':       (True,      deprecated),
    'window.editbar':       (True,      lambda v: type(v) is bool),
    'window.statusbar':     (True,      lambda v: type(v) is bool),
    #TODO: reintroduce it, but per game.type
    #'draw.default_rotation':((-30.,39.),lambda v: tuple_validator(2, 2, float, v)),
    #TODO: deprecated
    'draw.lighting':        (True,     lambda v: type(v) is bool),
    'draw.shader':          (0,         shaders(['lighting', 'simple', 'label'])),
    'draw.selection':       (1,         ['quadrant', 'simple', 'gesture']),
    'draw.speed':           (30,        (1, 100)),
    'game.type':            ('Cube',    validate_mtype),
    'game.size':            (None,      deprecated),
    'game.blocks':          (None,      deprecated),
    'game.moves':           (None,      deprecated),
    'game.position':        (None,      deprecated),
    'games.*.size':         (None,      lambda v: tuple_validator(0, 10, int, v, (1, 10))),
    'theme.face.*.color':       (None,          deprecated),
    'theme.face.*.image':       (None,          deprecated),
    'theme.face.*.mode':        (None,          deprecated),
    
    'theme.faces.Up.color':          (None, deprecated),
    'theme.faces.Down.color':        (None, deprecated),
    'theme.faces.Left.color':        (None, deprecated),
    'theme.faces.Right.color':       (None, deprecated),
    'theme.faces.Front.color':       (None, deprecated),
    'theme.faces.Back.color':        (None, deprecated),
    'theme.faces.Front_Left.color':  (None, deprecated),
    'theme.faces.Front_Right.color': (None, deprecated),
    'theme.faces.Back_Right.color':  (None, deprecated),
    'theme.faces.Back_Left.color':   (None, deprecated),
    'theme.faces.Up.image':          (None, deprecated),
    'theme.faces.Down.image':        (None, deprecated),
    'theme.faces.Left.image':        (None, deprecated),
    'theme.faces.Right.image':       (None, deprecated),
    'theme.faces.Front.image':       (None, deprecated),
    'theme.faces.Back.image':        (None, deprecated),
    'theme.faces.Front_Left.image':  (None, deprecated),
    'theme.faces.Front_Right.image': (None, deprecated),
    'theme.faces.Back_Right.image':  (None, deprecated),
    'theme.faces.Back_Left.image':   (None, deprecated),
    'theme.faces.Up.mode':           (None, deprecated),
    'theme.faces.Down.mode':         (None, deprecated),
    'theme.faces.Left.mode':         (None, deprecated),
    'theme.faces.Right.mode':        (None, deprecated),
    'theme.faces.Front.mode':        (None, deprecated),
    'theme.faces.Back.mode':         (None, deprecated),
    'theme.faces.Front_Left.mode':   (None, deprecated),
    'theme.faces.Front_Right.mode':  (None, deprecated),
    'theme.faces.Back_Right.mode':   (None, deprecated),
    'theme.faces.Back_Left.mode':    (None, deprecated),
    
    'theme.faces.up.color':     ('#a81407',     lambda v: type(v) is str), # rot
    'theme.faces.down.color':   ('#d94b1c',     lambda v: type(v) is str), # orange
    'theme.faces.left.color':   ('#e3e3e3',     lambda v: type(v) is str), # weiß
    'theme.faces.right.color':  ('#f0c829',     lambda v: type(v) is str), # gelb
    'theme.faces.front.color':  ('#1d6311',     lambda v: type(v) is str), # dunkelgrün
    'theme.faces.back.color':   ('#00275e',     lambda v: type(v) is str), # dunkelblau
    
    'theme.faces.front_left.color':      ('#65b5b3', lambda v: type(v) is str), # hellblau
    'theme.faces.front_right.color':     ('#f0c829', lambda v: type(v) is str), # gelb
    'theme.faces.back_right.color':      ('#771193', lambda v: type(v) is str), # violett
    'theme.faces.back_left.color':       ('#e3e3e3', lambda v: type(v) is str), # weiß
    
    'theme.faces.down_back_right.color': ('#00275e', lambda v: type(v) is str), # dunkelblau
    'theme.faces.down_front_right.color':('#646464', lambda v: type(v) is str), # grau
    'theme.faces.down_front.color':      ('#1d6311', lambda v: type(v) is str), # dunkelgrün
    'theme.faces.down_front_left.color': ('#f260b8', lambda v: type(v) is str), # pink
    'theme.faces.down_back_left.color':  ('#9b661b', lambda v: type(v) is str), # braun
    
    'theme.faces.up_back.color':         ('#86d624', lambda v: type(v) is str), # hellgrün
    'theme.faces.up_back_right.color':   ('#771193', lambda v: type(v) is str), # violett
    'theme.faces.up_front_right.color':  ('#f0c829', lambda v: type(v) is str), # gelb
    'theme.faces.up_front_left.color':   ('#65b5b3', lambda v: type(v) is str), # hellblau
    'theme.faces.up_back_left.color':    ('#e3e3e3', lambda v: type(v) is str), # weiß
    
    'theme.faces.*.image':      ('',            lambda v: type(v) is str),
    'theme.faces.*.mode':       (0,             ['tiled', 'mosaic']),
    'theme.bgcolor':            ('#7e9190',     lambda v: type(v) is str),
    'draw.accels':          ([('r', 'Num+6'), ('r-', 'Shift+Num+Right'),
                              ('l', 'Num+4'), ('l-', 'Shift+Num+Left'),
                              ('u', 'Num+8'), ('u-', 'Shift+Num+Up'),
                              ('d', 'Num+2'), ('d-', 'Shift+Num+Down'),
                              ('f', 'Num+5'), ('f-', 'Shift+Num+Clear'),
                              ('b', 'Num+0'), ('b-', 'Shift+Num+Ins'),
                              ('R', 'Ctrl+Num+8'), ('L', 'Ctrl+Num+2'),
                              ('U', 'Ctrl+Num+4'), ('D', 'Ctrl+Num+6'),
                              ('F', 'Ctrl+Num+5'), ('B', 'Ctrl+Num+0'),
                             ],         validate_draw_accels),
    'draw.zoom':            (1.4,       (0.1, 100.0)),
    #TRANSLATORS: The following 6 words are for the antialiasing levels: disabled, ugly, low, medium, high, higher
    'draw.samples':         (3,         [N_('disabled'), N_('ugly'), N_('low'),
                                         N_('medium'), N_('high'), N_('higher')]),
    'draw.mirror_faces':    (False,     lambda v: type(v) is bool),
    'draw.mirror_distance': (2.1,       (0.1, 10.0)),
    'action.edit_moves':    ('Ctrl+L',  lambda v: type(v) is str),
    'action.edit_cube':     ('',        lambda v: type(v) is str),
    'action.selectmodel':   ('Ctrl+M',  lambda v: type(v) is str),
    'action.initial_state': ('',        lambda v: type(v) is str),
    'action.reset_rotation':('Ctrl+R',  lambda v: type(v) is str),
    'action.preferences':   ('Ctrl+P',  lambda v: type(v) is str),
   }




#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os

N_ = lambda s: s
try:
    _
except NameError:
    _ = N_

AUTHOR = 'B. Clausius'
CONTACT_EMAIL = 'barcc@gmx.de'
CONTACT_WEB = 'https://answers.launchpad.net/pybik'
COPYRIGHT = 'Copyright © 2009-2017,  B. Clausius'
LICENSE_NAME = 'GPL-3.0+'
PACKAGE = 'pybik'
PACKAGE_FULL = 'pybik.barcc'
#TRANSLATORS: Name of the application, probably should not be translated.
APPNAME = N_('Pybik')
VERSION = '3.0'
VERSION_SHORT = VERSION.split('~')[0]
WEBSITE = 'https://launchpad.net/pybik/'
DOWNLOADS = 'https://launchpad.net/pybik/+download'
BUG_DATABASE = 'https://bugs.launchpad.net/pybik'
CONTACT_FILEBUG = BUG_DATABASE + '/+filebug'


# The following two lines are replaced by "setup.py install"
data_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data')
appdata_dir = data_dir
# The following line is replaced by "setup.py sdist"
RELEASE_DATE = '2017-10-22'

LOCALE_DIR   = os.path.join(data_dir, 'locale')
APPICON_FILE = os.path.join(appdata_dir, 'icons', 'pybik-64.png')
APPICON192_FILE = os.path.join(appdata_dir, 'icons', 'pybik-192.png')
LICENSE_FILE = os.path.join(appdata_dir, 'GPL-3')
MODELS_DIR   = os.path.join(appdata_dir, 'models')
PLUGINS_DIR  = os.path.join(appdata_dir, 'plugins')
UI_DIR       = os.path.join(appdata_dir, 'ui')
SHADER_DIR   = os.path.join(appdata_dir, 'shaders')

def get_config_home(*args):
    return os.path.join(os.environ.get('XDG_CONFIG_HOME', '') or os.path.expanduser("~/.config"), *args)
def get_data_home(*args):
    return os.path.join(os.environ.get('XDG_DATA_HOME', '') or os.path.expanduser("~/.local/share"), *args)
    
#TRANSLATORS: short description
SHORT_DESCRIPTION = N_("Rubik's cube game")
LONG_DESCRIPTION = (
    #TRANSLATORS: The next 7 lines belong together and form the long description
    N_('Pybik is a 3D puzzle game about the cube invented by Ernő Rubik.'),
    '',
    '* '+N_('Various 3D puzzles, such as cubes, towers, prisms, tetrahedra, octahedra, dodecahedra and many more'),
    '* '+N_('Solvers for some puzzles'),
    '* '+N_('Pretty patterns'),
    '* '+N_('Editor for move sequences'),
    '* '+N_('Custom colors and images on the surface of the puzzles'),
)

LICENSE_INFO = N_(
            'This program is free software: you can redistribute it and/or modify '
            'it under the terms of the GNU General Public License as published by '
            'the Free Software Foundation, either version 3 of the License, or '
            '(at your option) any later version.\n\n'

            'This program is distributed in the hope that it will be useful, '
            'but WITHOUT ANY WARRANTY; without even the implied warranty of '
            'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the '
            'GNU General Public License for more details.')
LICENSE_FURTHER = N_(
#TRANSLATORS: Text between "<" and ">" is expanded to a link by the program and should not be modified.
#             Text between "<LINK|>" and "<|>" is the translatable text for the link.
            'Read the full text of the <text:FULL_LICENSE_TEXT|>GNU General Public License'
            '<|> or see <http://www.gnu.org/licenses/>.')
LICENSE_NOT_FOUND = N_(
            'You should have received a copy of the GNU General Public License'
            ' along with this program.  If not, see <http://www.gnu.org/licenses/>.')

def get_filebug_text():
    return _(
    'If you find any bugs in Pybik or have a suggestion for an improvement then please '
    'submit a <{CONTACT_FILEBUG}|>bug report<|>.'
    ).format(CONTACT_FILEBUG=CONTACT_FILEBUG)

TRANSLATION_TEXT = N_(
    'Translations are managed by the '
    '<https://translations.launchpad.net/+groups/launchpad-translators|>'
    'Launchpad translation group<|>.\n\n'
    'If you want help to translate Pybik to your language you can do it through '
    'the <https://translations.launchpad.net/pybik|>web interface<|>.\n\n'
    'Read more about <https://help.launchpad.net/Translations|>"Translating with Launchpad"<|> '
    'and <https://help.launchpad.net/Translations/StartingToTranslate|>"Starting to translate"<|>.')

REPOSITORY_SHORT = 'lp:pybik'
REPOSITORY_URL = 'https://code.launchpad.net/~barcc/pybik/trunk'
REPOSITORY_BROWSE = 'https://bazaar.launchpad.net/~barcc/pybik/trunk/files'
REPOSITORY_DOWNLOAD = 'http://bazaar.launchpad.net/~barcc/pybik/trunk/download'





#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os

from . import debug, config

N_ = lambda s: s


class Tee2(object):
    class Tee(object):
        def __init__(self, file1, file2):
            self.file1 = file1
            self.file2 = file2
        def write(self, data):
            self.file1.write(data)
            self.file1.flush()
            self.file2.write(data)
        def flush(self):
            self.file1.flush()
            self.file2.flush()
            
    def __init__(self, name='pybik.log', mode='w'):
        self.file = open(name, mode)
        self.file.write('logfile created\n')
        self.file.flush()
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self.Tee(self.file, self.stdout)
        sys.stderr = self.Tee(self.file, self.stderr)
        
    def __del__(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        self.file.close()
        
        
class Options:
    test_arg_names = ['write-y', 'write-yes', 'write-n', 'write-no', 'write-e', 'write-error',
                      'log-widgets', 'notime']
    arg_info = [
        (None, None, '\n'.join(config.LONG_DESCRIPTION)),
        (None, None, 'Pybik has a graphical user interface.'
                     ' The options are intended for testing and debugging.'),
        (None, None, 'Options:'),
        (['-h', '--help'], lambda self, value: self.print_usage(),
                            'Show help message and exit'),
        (['--version'], lambda self, value: self.print_version(),
                            'Show version number and exit'),
        (['--config-file='], lambda self, value: self.__dict__.update(config_file=value),
                            'Specify the configuration filename'),
        (['--defaultconfig'], lambda self, value: self.__dict__.update(defaultconfig=True),
                            'Print default settings to stdout and exit'),
        (['--games-file='], lambda self, value: self.__dict__.update(games_file=value),
                            'Specify the filename for saved games'),
        (['--qml-file='], lambda self, value: self.__dict__.update(qml_file=value),
                            'Alternate user interface'),
        (['--variant='], lambda self, value: self.__dict__.update(variant=value),
                            'Extension module variant'),
        (['--demo='], lambda self, value: self.__dict__.update(demo=value),
                            'Demo module'),
        (['--log'], lambda self, value: self.tee(),
                            'Write logfile'),
        #TODO: test options removed in revision 4196
        (['--debug='], lambda self, value: debug.set_flagstring(value),
                            'D is a comma-separated list of debug flags:\n{0}'
                            .format(' '.join(debug.flag_nicks))),
        (None, None,  'Qt-Options, for a full list refer to the Qt-Reference,'
                ' but not all options make sense for this application:'),
        (['-qwindowgeometry G', '-geometry G'], None,
                            'Specifies window geometry for the main window using the X11-syntax.'
                            ' For example: 100x100+50+50'),
        (['-reverse'], None,
                            'Sets the application layout direction to left-to-right'),
        (['-widgetcount'], None,
                            'Prints debug message at the end about number of widgets left'
                            ' undestroyed and maximum number of widgets existed at the same time'),
        ]
        
    def __init__(self):
        self.config_file = None
        self.defaultconfig = False
        self.games_file = None
        self.qml_file = None
        self.variant = None
        self.demo = None
        self.pure_python = False
        self.test = []
        self.test_args = []
        self.ui_args = []
        
    @staticmethod
    def format_opts(args):
        args = ((a + a.strip('--')[0].upper() if a.endswith('=') else a) for a in args)
        return '  ' + ', '.join(args)
        
    @staticmethod
    def format_help(text, indent=0):
        try:
            width = int(os.environ['COLUMNS']) - 2
        except (KeyError, ValueError):
            width = 78
        width -= indent
        def split(text):
            lines = text.split('\n')
            for line in lines:
                res = None
                words = line.split(' ')
                for word in words:
                    if res is None:
                        res = word
                    elif len(res + word) + 1 <= width:
                        res = ' '.join((res, word))
                    else:
                        yield res
                        res = word
                yield res
        return '\n'.ljust(indent+1).join(split(text))
        
    @classmethod
    def print_usage(cls):
        print('Usage:', sys.argv[0], '[options]')
        for args, unused_func, text in cls.arg_info:
            if args is None:
                print()
                print(cls.format_help(text))
            else:
                opts = cls.format_opts(args)
                if len(opts) > 18:
                    text = '\n' + text
                else:
                    opts = opts.ljust(20)
                text = cls.format_help(text, 20)
                print(opts + text)
        sys.exit(0)
        
    @staticmethod
    def print_version():
        from textwrap import fill
        print(config.APPNAME, config.VERSION)
        print()
        print(config.COPYRIGHT)
        print()
        print('\n\n'.join(fill(text, width=78) for text in config.LICENSE_INFO.split('\n\n')))
        print()
        print(fill(config.LICENSE_NOT_FOUND, width=78))
        sys.exit(0)
        
    def tee(self):
        self._tee = Tee2()
        
    def parse(self, args):
        arg_functs = {o: f for opts, f, h in self.arg_info for o in opts or [] if f is not None}
        for arg in args:
            try:
                index = arg.index('=')
            except ValueError:
                key, value = arg, None
            else:
                key, value = arg[:index+1], arg[index+1:]
            try:
                func = arg_functs[key]
            except KeyError:
                self.ui_args.append(arg)
            else:
                func(self, value)
                
        for a in self.test_args:
            if a not in self.test_arg_names:
                print('unknown test argument:', a)
                sys.exit(1)
                
        
class Main:
    def __init__(self, root_dir=None):
        self.root_dir = root_dir
        self.opts = Options()
        
        self.opts.parse(sys.argv)
        
        from .debug import DEBUG, DEBUG_QSGINFO, DEBUG_BASICRENDERER, DEBUG_PUREPYTHON, DEBUG_PHONE, DEBUG_CLOCALE
        if DEBUG:
            print('Qt args:', self.opts.ui_args)
            if DEBUG_QSGINFO:
                print('QSG_INFO=1')
                os.environ['QSG_INFO'] = '1'
            if DEBUG_BASICRENDERER:
                print('QSG_RENDER_LOOP=basic')
                os.environ['QSG_RENDER_LOOP'] = 'basic'
            if DEBUG_CLOCALE:
                os.environ['LANGUAGE'] = ''
                os.environ['LC_ALL'] = 'C.UTF-8'
                print('LC_ALL='+os.environ['LC_ALL'])
        
        if self.opts.defaultconfig:
            from .settings import settings
            settings.load('')
            settings.dump(sys.stdout, all=True)
            sys.exit(0)
            
        package = config.PACKAGE_FULL if DEBUG_PHONE else config.PACKAGE
        if self.opts.config_file is None:
            self.opts.config_file = config.get_config_home(package, 'settings.conf')
        elif not self.opts.config_file:
            self.opts.config_file = None
        if self.opts.games_file is None:
            self.opts.games_file = config.get_data_home(package, 'games')
        elif not self.opts.games_file:
            self.opts.games_file = None
        if self.opts.demo:
            if self.opts.games_file is not None:
                print('saved games not used in demo mode')
                self.opts.games_file = None
            if self.opts.config_file is not None:
                print('saved config not used in demo mode')
                self.opts.config_file = None
        if self.opts.qml_file is None:
            self.opts.qml_file = os.path.join(config.UI_DIR, 'qt', 'main.qml')
            
        #TODO: purepython mode currently not implemented
        if DEBUG_PUREPYTHON and False:
            from .ext import qtexec
        else:
            from .ext import _qtexec_ as qtexec
        self.qtexec = qtexec
            
        res = qtexec.exec_application(self)
        if res:
            sys.exit(res)
            
    def resolve_variant(self):
        from .debug import DEBUG, DEBUG_IMPORT
        if self.opts.variant is None:
            if DEBUG_IMPORT: print('no variant specified, using default')
            self.opts.variant = self.qtexec.aliases['']
        elif self.opts.variant not in self.qtexec.depends:
            if self.opts.variant in self.qtexec.aliases:
                if DEBUG_IMPORT: print('alias variant specified:', repr(self.opts.variant))
                self.opts.variant = self.qtexec.aliases[self.opts.variant]
            else:
                if DEBUG_IMPORT: print('unknown variant: ', repr(self.opts.variant), ', using default instead', sep='')
                self.opts.variant = self.qtexec.aliases['']
        if '???' in self.opts.variant:
            if DEBUG_IMPORT: print('autodetect GL-variant')
            self.opts.variant = self.opts.variant.replace('???', self.qtexec.get_gl_variant())
        if DEBUG: print('variant:', self.opts.variant)
        assert self.opts.variant in self.qtexec.depends
        
    def load_gllib(self):
        from .debug import DEBUG_IMPORT
        #XXX: Workaround for lp:941826, "dlopen(libGL.so) resolves to mesa rather than nvidia"
        #     Was: "PyQt cannot compile shaders with Ubuntu's Nvidia drivers"
        #     https://bugs.launchpad.net/ubuntu/+source/python-qt4/+bug/941826
        #XXX: Also needed in compiled mode. Without this, on the phone the OpenGL entry points
        #     in the first module are NULL and the app will crash
        # The Python doc says:
        #   it may be better to determine the shared library name at development time,
        #   and hardcode that into the wrapper module instead of using find_library()
        #   to locate the library at runtime.
        # find_library() also may fail as it relies on external tools, so hardcode the names
        import ctypes
        # self.opts.variant must be a full variant name: (qtw|qtq)(ogl|es2)[os][d]
        gllib = {'ogl': 'libGL.so.1', 'es2': 'libGLESv2.so.2'}[self.opts.variant[3:6]]
        if DEBUG_IMPORT:
            import ctypes.util
            glname = gllib.split('.')[0][3:]
            print('OpenGL library:', gllib, glname, ctypes.util.find_library(glname))
        try:
            ctypes.CDLL(gllib, ctypes.RTLD_GLOBAL)
        except OSError as e:
            import ctypes.util
            print('OpenGL library not found:', gllib, glname)
            print('   ', e)
            glname = gllib.split('.')[0][3:]
            gllib = ctypes.util.find_library(glname)
            if gllib is None:
                print('    not found by name {!r}, continue without preloading'.format(glname))
            else:
                print('    retry with:', gllib)
                try:
                    ctypes.CDLL(gllib, ctypes.RTLD_GLOBAL)
                except OSError as e:
                    print('    not found, continue without preloading:')
                    print('   ', e)
        
    def cb_app_post_create(self, language):
        self.resolve_variant()
        self.load_gllib()
        
        from .debug import DEBUG_IMPORT
        import pybiklib.ext
        for extname in self.qtexec.depends[self.opts.variant]:
            name = extname.split('_')[1]
            if DEBUG_IMPORT:
                print('importing extension', extname, 'as', name)
            __import__('pybiklib.ext', fromlist=(extname,))
            module = getattr(pybiklib.ext, extname)
            sys.modules['pybiklib.ext.'+name] = module
            setattr(pybiklib.ext, name, module)
            if debug is not None:
                module.set_debug_flags(debug)
            
        import gettext
        if self.root_dir == sys.prefix:
            # normal installation
            LOCALEDIR = None
        else:
            # different root, e.g. /usr/local, source directory
            LOCALEDIR = config.LOCALE_DIR
        t = gettext.translation(config.PACKAGE, LOCALEDIR, languages=[language], fallback=True)
        t.install(names=['ngettext'])
        
        from pybiklib.ext import qt
        qt.app_post_create(_)
        
    def cb_run_app(self):
        from .application import MainWindow
        if self.opts is None:
            return None
        elif self.opts.test:
            #TODO: removed in revision 4196
            return None
        else:
            window = MainWindow(self.opts)
            self.opts = None
            return window
            
    




#-*- coding:utf-8 -*-

#  Copyright © 2009-2015, 2017  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import random
from copy import deepcopy

from .debug import DEBUG_MSG, DEBUG_RAND, DEBUG_ROTATE
from .moves import MoveQueuePacked, MoveData


# This class is currently used to store the initial state of the cube
# and to provide the cube to plugins.
class CubeState:
    rand = random.Random()
    if DEBUG_RAND:
        print('rand.seed(123)')
        rand.seed(123)
    
    def __init__(self, model):
        self.model = model
        self._blocksn = []      # list of (block-position, rotation-symbol)
        self._blocksr = []      # list of (block-index, rotation-symbol)
        # if   self._blocksn[idx] == (pos, rot)
        # then self._blocksr[pos] == (idx, rot)
        
    def copy(self):
        other = CubeState(self.model)
        other._blocksn = deepcopy(self._blocksn)
        other._blocksr = deepcopy(self._blocksr)
        return other
        
    @property
    def rotations(self):
        return [(idx, self.model.rotation_indices[rot], rot)
                    for pos, (idx, rot) in enumerate(self._blocksr)]
        
    def set_solved(self):
        # '' symbolic rotation for identity
        self._blocksn = [(i, '') for i in range(self.model.cell_count)]
        self._blocksr = [(i, '') for i in range(self.model.cell_count)]
        
    def is_solved(self):
        '''Check whether the cube is solved.
        
        Only test the colors of the visible faces.
        Cubes with rotated center faces are treated as solved.
        '''
        face_colors = {}
        for blockpos, cells_visible_faces in enumerate(self.model.cells_visible_faces):
            for face in cells_visible_faces:
                color = self.get_colorsymbol(blockpos, face)
                try:
                    color_ref = face_colors[face]
                except KeyError:
                    face_colors[face] = color
                else:
                    if color != color_ref:
                        return False
        return True
        
    def is_solved_strict(self):
        '''Check whether the cube is solved.
        
        Test the rotation and position of the blocks.
        Cubes with rotated center faces are treated as not solved.
        '''
        #FIXME: currently unused function, should be used for faces with images
        allrot = self._blocksn[0][1]
        for index, (pos, rot) in enumerate(self._blocksn):
            if rot != allrot:
                return False  # Cubie rotated
            block = self.model.rotated_position[rot][index]
            if pos != block:
                return False  # Cubie at wrong place
        return True
        
    def is_identity(self):
        for index, (pos, rot) in enumerate(self._blocksn):
            if index != pos or rot != '':
                return False
        return True
        
    def identify_rotation_blocks(self, maxis, mslice):
        if mslice == -1:
            for i, blockn in enumerate(self._blocksn):
                yield i, blockn
        else:
            for i, blockn in enumerate(self._blocksn):
                bslice = self.model.cell_indices[blockn[0]][maxis]
                if bslice == mslice:
                    yield i, blockn
                
    def _rotate_slice(self, axis, slice_, dir_):
        if DEBUG_ROTATE:
            print('rotate axis={} slice={} dir={!s:5}\n  blocks:'.format(axis, slice_, dir_), end=' ')
        for idx, (pos, rot) in self.identify_rotation_blocks(axis, slice_):
            if DEBUG_ROTATE:
                print('{}:{}{}'.format(idx, pos, rot), end='')
            pos, rot = self.model.rotate_symbolic(axis, dir_, pos, rot)
            if DEBUG_ROTATE:
                print('-{}{}'.format(pos, rot), end=' ')
            self._blocksn[idx] = pos, rot
            self._blocksr[pos] = idx, rot
        if DEBUG_ROTATE:
            print()
        
    def get_colorsymbol(self, blockpos, facesym):
        rot = self._blocksr[blockpos][1]
        return self.model.face_symbolic_to_face_color(facesym, rot)
        
    def get_colornum(self, blockpos, facesym):
        colorsym = self.get_colorsymbol(blockpos, facesym)
        return self.model.faces.index(colorsym)
        
    def format_block(self):
        # every block is stored as pos-sym, where sym is a symbolic rotation
        blocks = ['{}-{}'.format(pos, sym) for pos, sym in self._blocksn]
        return 'idx-rot: ' + ' '.join(blocks)
        
    __str__ = format_block
    
    def parse_block(self, blocks):
        if blocks == 'solved':
            return self.set_solved()
        bformat, blocks = blocks.split(':', 1)
        if bformat != 'idx-rot':
            raise ValueError('unknown block format:', bformat)
        blocks = blocks.strip().split(' ')
        if len(blocks) != self.model.cell_count:
            raise ValueError('wrong block count: %s, expected: %s' % (len(blocks), self.model.cell_count))
        blocksn = []
        for block in blocks:
            # every block is stored as idx-rot, where idx: index to blocklist, rot: symbolic rotation
            block = block.strip().split('-', 1)
            index, rot = block
            index = int(index)
            rot = self.model.norm_symbol(rot)
            blocksn.append((index, rot))
        # test whether block indices is a permutation,
        # in fact thats not enough, e.g. swap a corner cubie with an edge,
        # also cubie rotation may be invalid, it can be possible that a
        # label is rotated inside the cube.
        for i1, i2 in enumerate(sorted(i for i, r in blocksn)):
            if i1 != i2:
                raise ValueError('block list is not a permutation')
        self._blocksn = blocksn
        self._blocksr = [None] * len(blocksn)
        for i, (pos, rot) in enumerate(blocksn):
            self._blocksr[pos] = (i, rot)
        
    def random(self, count=-1):
        sizes = self.model.sizes
        all_moves = [(a,s,d) for a,_sz in enumerate(sizes) for s in range(_sz) for d in (False,True)]
        randrange = self.rand.randrange
        if count < 0:
            count = 10 * len(all_moves) + randrange(len(all_moves))
        if count:
            # The solved cube should be randomly rotated (lp: 1411999)
            for sym in self.rand.choice(self.model.rotation_symbols):
                self.rotate_slice(self.model.sym_to_move(sym))
        packed = MoveQueuePacked()
        all_moves = [(a,s,d) for a,s,d in all_moves if self.model.is_user_slice(a, s)]
        for i in range(count):
            candidates = list(range(len(all_moves)))
            while candidates:
                maxis, mslice, mdir = all_moves[candidates.pop(randrange(len(candidates)))]
                packed.push_unpacked(MoveData(maxis, mslice, mdir), self.model)
                if packed.length_no_full(self.model) > i:
                    self._rotate_slice(maxis, mslice, mdir)
                    break
                packed.push_unpacked(MoveData(maxis, mslice, not mdir), self.model)
            else:
                break
        if DEBUG_MSG:
            print(count, 'random moves:', packed.format(self.model)[0])
            
    def rotate_slice(self, move_data):
        self._rotate_slice(*move_data)
        
    def swap_block(self, blockpos1, maxis, mslice, mdir):
        if DEBUG_ROTATE:
            print('rotate axis={} slice={} dir={!s:5}\n  from:'.format(maxis, mslice, mdir), end=' ')
            
        blockidx1, blockrot1 = self._blocksr[blockpos1]
        blockpos2, blockrot1r = self.model.rotate_symbolic(maxis, mdir, blockpos1, blockrot1)
        blockidx2, blockrot2 = self._blocksr[blockpos2]
        blockpos2r, blockrot2r = self.model.rotate_symbolic(maxis, not mdir, blockpos2, blockrot2)
        if DEBUG_ROTATE:
            print('{}:{}{}->{}{}\n    to: {}:{}{}->{}{}'.format(
                blockidx1, blockpos1, blockrot1, blockpos2, blockrot1r,
                blockidx2, blockpos2, blockrot2, blockpos2r, blockrot2r))
        assert blockpos2r == blockpos1
        self._blocksn[blockidx1] = blockpos2, blockrot1r
        self._blocksr[blockpos2] = blockidx1, blockrot1r
        self._blocksn[blockidx2] = blockpos2r, blockrot2r
        self._blocksr[blockpos2r] = blockidx2, blockrot2r
        
    def rotate_block(self, blockpos, rdir):
        blockidx, blockrot = self._blocksr[blockpos]
        try:
            rot = self.model.inplace_rotations[blockpos][-1 if rdir else 0]
        except IndexError:
            # not every block can be rotated inline: e.g. edges and center faces on the 4×4×4-Cube,
            # edges and corners on towers and bricks
            #TODO: swap edge at pos n with the one at pos (size-1 - n),
            #      rotate all center faces on the same ring
            return
        blockrot2 = self.model.norm_symbol(blockrot + rot)
        self._blocksn[blockidx] = blockpos, blockrot2
        self._blocksr[blockpos] = blockidx, blockrot2
        if DEBUG_ROTATE:
            sym1, colorsym1 = self.model.blockpos_to_blocksym(blockpos, blockrot)
            sym2, colorsym2 = self.model.blockpos_to_blocksym(blockpos, blockrot2)
            print('{}:{}{}->{}{} ({}:{}->{}:{})'.format(blockidx, blockpos, blockrot, blockpos, blockrot2,
                                                      sym1, colorsym1, sym2, colorsym2))
            self.debug_blocksymbols(allsyms=False)
            
    def debug_blocksymbols(self, allsyms):
        for blockpos, blockrot in self._blocksn:
            blocksym, colorsym = self.model.blockpos_to_blocksym(blockpos, blockrot)
            if allsyms or blocksym != colorsym:
                print(' {}:{}'.format(blocksym, colorsym), end='')
        print('')
        
        
